HORIZONTAL = ["left", "center", "right"]
VERTICAL = ["top", "middle", "bottom"]


def position(*args):
    """Returns a formatted position value as string.

    :Examples:
        >>> from imglab import position
        >>> position("left", "bottom")
        'left,bottom'
        >>> position("bottom", "left")
        'bottom,left'
        >>> position("left")
        'left'
        >>> position("bottom")
        'bottom'

    :param args: The position with two directions or one single direction as strings.
    :type args: list
    :return: A string value representing the color or a `ValueError` exception if the arguments are invalid.
    :rtype: str
    """
    if len(args) == 1 and __valid_position(args[0]):
        return args[0]
    elif len(args) == 2 and __valid_position(*args):
        return ",".join(args)
    else:
        raise ValueError("Invalid position")


def __valid_position(*directions):
    if len(directions) == 1 and __valid_direction(directions[0]):
        return True
    elif len(directions) == 2 and __valid_directions(directions[0], directions[1]):
        return True
    else:
        return False


def __valid_direction(direction):
    return direction in HORIZONTAL or direction in VERTICAL


def __valid_directions(direction_a, direction_b):
    return (direction_a in HORIZONTAL and direction_b in VERTICAL) or (
        direction_b in HORIZONTAL and direction_a in VERTICAL
    )



from time import struct_time
from datetime import datetime
from calendar import timegm
from urllib.parse import urlparse

WEB_URI_SCHEMES = ["https", "http"]


def normalize_path(path):
    """Returns a normalized path where suffix and prefix slashes are removed.

    :Examples:
        >>> from imglab import utils
        >>> utils.normalize_path("/example.jpeg")
        'example.jpeg'
        >>> utils.normalize_path("/example.jpeg/")
        'example.jpeg'
        >>> utils.normalize_path("/subfolder/example.jpeg/")
        'subfolder/example.jpeg'

    :param path: A path string to be normalized.
    :type path: str
    :return: A string with the normalized path.
    :rtype: str
    """
    return path.strip("/")


def normalize_params(params):
    """Returns a dict of normalized params, transforming keys with underscores to hyphens and expires values with
    struct_time or datetime instances to timestamps.

    :Examples:
        >>> from imglab import utils
        >>> utils.normalize_params({"trim": "color", "trim_color": "orange"})
        {'trim': 'color', 'trim-color': 'orange'}
        >>> import time
        >>> utils.normalize_params({"width": 200, "height": 300, "expires": time.gmtime(1464096368)})
        {'width': 200, 'height': 300, 'expires': 1464096368}

    :param params: A dict with params to be normalized.
    :type params: dict
    :return: A dict with the normalized params.
    :rtype: dict
    """
    return dict(_normalize_param(_dasherize(key), value) for key, value in params.items())


def is_web_uri(uri):
    """Returns if the specified uri is a valid Web URI with https or https schema or not.

    :Examples:
        >>> from imglab import utils
        >>> utils.is_web_uri("https://imglab.io")
        True
        >>> utils.is_web_uri("http://imglab.io")
        True
        >>> utils.is_web_uri("imglab.io")
        False
        >>> utils.is_web_uri("ftp://imglab.io")
        False

    :param uri: A string with the URI.
    :type uri: str
    :return: `True` if the specified uri is a valid Web URI, `False` otherwise.
    :rtype: bool
    """
    try:
        return urlparse(uri).scheme in WEB_URI_SCHEMES
    except Exception:
        return False


def _dasherize(value):
    return value.replace("_", "-")


def _normalize_param(key, value):
    if key == "expires" and isinstance(value, struct_time):
        return (key, timegm(value))
    elif key == "expires" and isinstance(value, datetime):
        return (key, timegm(value.timetuple()))
    else:
        return (key, value)



import base64
import hmac
import hashlib


def generate(source, path, encoded_params=None):
    """Returns a generated signature for a source, path and encoded parameters.

    :Examples:
        >>> from imglab import signature, Source
        >>> source = Source("assets", secure_key="55IX1RVlDHpgl/4D", secure_salt="ITvYA2lPfyz0w8/v")
        >>> signature.generate(source, "example.jpeg")
        'QFEVlDWgK289HYKr2KJdwtPC-I7LS195hSVQhS1UsRA'

    :param source: The source used to generate the signature.
    :type source: class:`imglab.Source`
    :param path: The path of the resource.
    :type path: str
    :param encoded_params: Encoded query params of the URL to generate the signature.
    :type encoded_params: str, optional
    :return: A string with the signature encoded using Base64 to be used in a imglab URL.
    :rtype: str
    """
    decoded_secure_key = base64.b64decode(source.secure_key.encode())
    decoded_secure_salt = base64.b64decode(source.secure_salt.encode())

    data = b"%s/%s" % (decoded_secure_salt, path.encode())
    data = b"%s?%s" % (data, encoded_params.encode()) if encoded_params else data

    digest = hmac.new(decoded_secure_key, data, hashlib.sha256).digest()

    return base64.urlsafe_b64encode(digest).rstrip(b"=").decode()



from urllib.parse import ParseResult, quote, urlencode

from .source import Source
from .signature import generate as generate_signature
from .utils import normalize_path, normalize_params, is_web_uri


def url(source, path, **params):
    """
    Returns a formatted URL string for a source, with a path and some optional specified arguments.

    :Examples:
        >>> import imglab
        >>> imglab.url("assets", "example.jpeg")
        'https://assets.imglab-cdn.net/example.jpeg'
        >>> imglab.url(imglab.Source("assets"), "example.jpeg", width=500, height=600)
        'https://assets.imglab-cdn.net/example.jpeg?width=500&height=600'

    :param source: A source name as string or `imglab.Source` class instance.
    :type source: str, class:`imglab.Source`
    :param path: The path where the resource is located.
    :type path: str
    :param params: The query parameters that we want to use as a keyword argument list.
    :type params: list, optional
    :return: A string with the formatted URL with the specified arguments.
    :rtype: str
    """
    if isinstance(source, str):
        return _url_for_source(Source(source), path, params)
    elif isinstance(source, Source):
        return _url_for_source(source, path, params)
    else:
        raise ValueError("Invalid source name or source. A string or a %s instance is expected." % Source.__name__)


def _url_for_source(source, path, params):
    normalized_path = normalize_path(path)
    normalized_params = normalize_params(params)

    return ParseResult(
        scheme=source.scheme(),
        netloc=_netloc(source),
        path=source.path(_encode_path(normalized_path)),
        params=None,
        query=_encode_params(source, normalized_path, normalized_params),
        fragment=None,
    ).geturl()


def _netloc(source):
    if source.port:
        return ":".join([source.host, str(source.port)])
    else:
        return source.host


def _encode_path(path):
    if is_web_uri(path):
        return _encode_path_component(path)
    else:
        return "/".join(map(_encode_path_component, path.split("/")))


def _encode_path_component(path_component):
    return quote(path_component, safe=[])


def _encode_params(source, path, params):
    if source.is_secure():
        params.update(signature=generate_signature(source, path, urlencode(params)))

    return urlencode(params)



import os


class Source:
    """A class to represent imglab sources.

    :Examples:
        >>> import imglab
        >>> source = imglab.Source("assets")
        >>> source = imglab.Source("assets", subdomains=False)
        >>> source = imglab.Source("assets", https=False, host="imglab.net", port=8080)

    :param name: The name of source
    :type name: str
    :param host: A host to be used with the source
    :type host: str, optional
    :param https: A bool value specifying if the source should use https schema to generate host value
    :type https: bool, optional
    :param port: A integer identifying the port to be used with the source
    :type port: int, optional
    :param secure_key: A string value with the secure_key to be used with the source
    :type secure_key: str, optional
    :param secure_salt: A string value with the secure_salt to be used with the source
    :type secure_salt: str, optional
    :param subdomains: A bool value specifying if the source should use subdomains or not
    :type subdomains: bool, optional
    """

    DEFAULT_HOST = "imglab-cdn.net"
    DEFAULT_HTTPS = True
    DEFAULT_SUBDOMAINS = True

    def __init__(
        self,
        name,
        host=DEFAULT_HOST,
        https=DEFAULT_HTTPS,
        port=None,
        secure_key=None,
        secure_salt=None,
        subdomains=DEFAULT_SUBDOMAINS,
    ):
        self._host = host
        self._https = https
        self._name = name
        self._port = port
        self._secure_key = secure_key
        self._secure_salt = secure_salt
        self._subdomains = subdomains

    @property
    def host(self):
        """Returns the host used by the source.

        :return: A string value with the host used by the source.
        :rtype: str
        """
        if self.subdomains:
            return "%s.%s" % (self.name, self._host)
        else:
            return self._host

    @property
    def https(self):
        """Returns if the source uses https or not.

        :return: `True` if the source uses https scheme, `False` otherwise.
        :rtype: bool
        """
        return self._https

    @property
    def name(self):
        """Returns the name of the source.

        :return: A string value with the name of the source.
        :rtype: str
        """
        return self._name

    @property
    def port(self):
        """Returns the port used by the source.

        :return: An integer value identifying the port used by the source or `None` if no port is used.
        :rtype: int, None
        """
        return self._port

    @property
    def secure_key(self):
        """Returns the secure_key used by the source.

        :return: A string value with the secure_key used by the source or `None` if no secure_key is used.
        :rtype: str, None
        """
        return self._secure_key

    @property
    def secure_salt(self):
        """Returns the secure_salt used by the source.

        :return: A string value with the secure_salt used by the source or `None` if no secure_salt is used.
        :rtype: str, None
        """
        return self._secure_salt

    @property
    def subdomains(self):
        """Returns if the source uses subdomains or not.

        :return: `True` if the source uses subdomains, `False` otherwise.
        :rtype: bool
        """
        return self._subdomains

    def scheme(self):
        """Returns the URI scheme to be used with the source ("http" or "https").

        :return: `https` if the source is using https scheme, `http` otherwise.
        :rtype: str
        """
        if self.https:
            return "https"
        else:
            return "http"

    def path(self, path):
        """Returns the path to be used with the source.

        :param path: A path to be used.
        :type path: str
        :return: A string with the path to be used with the source.
        :rtype: str
        """
        if self.subdomains:
            return path
        else:
            return os.path.join(self.name, path)

    def is_secure(self):
        """Returns if the source is secure or not.

        :return: `True` if the source is secure, `False` otherwise.
        :rtype: bool
        """
        return bool(self.secure_key and self.secure_salt)



from .source import Source
from .color import color
from .position import position
from .url import url

from . import _version

__version__ = _version.version

__all__ = ["Source", "color", "position", "url"]
